r"""Base modules"""

from __future__ import annotations

__all__ = [
    'Module',
    'ModuleDef',
    'Variable',
    'Parameter',
    'Statistic',
]

import jax
import jax.tree_util as jtu

from jax import Array
from typing import *

from ..tree_util import PyArray, Static, Auto


def is_module(x: Any) -> bool:
    return isinstance(x, Module)


def is_parameter(x: Any) -> bool:
    return isinstance(x, Parameter)


class Module(Auto):
    r"""Base class for all modules.

    A module is a PyTree whose branches are its attributes. A branch can be any
    PyTree-compatible object (:py:`tuple`, :py:`list`, :py:`dict`, ...), including other
    modules. Parametric functions, such as neural networks, should subclass
    :class:`Module` and indicate their parameters with :class:`Parameter`.

    .. code-block:: python

        import jax
        import jax.random as jrd
        import inox
        import inox.nn as nn

        class Linear(nn.Module):
            def __init__(self, key, in_features, out_features):
                keys = jrd.split(key, 2)

                self.weight = Parameter(jrd.normal(keys[0], (in_features, out_features)))
                self.bias = Parameter(jrd.normal(keys[1], (out_features,)))

            def __call__(self, x):
                return x @ self.weight + self.bias

        class Classifier(nn.Module):
            def __init__(self, key, in_features, num_classes):
                keys = jrd.split(key, 3)

                self.l1 = Linear(keys[0], in_features, 64)
                self.l2 = Linear(keys[1], 64, 64)
                self.l3 = Linear(keys[2], 64, num_classes)
                self.relu = nn.ReLU()

                self.return_logits = True  # static leaf

            def __call__(self, x):
                x = self.l1(x)
                x = self.l2(self.relu(x))
                x = self.l3(self.relu(x))

                if self.return_logits:
                    return x
                else:
                    return jax.nn.softmax(x)

        key = jax.random.key(0)
        model = Classifier(key)

    Modules automatically detect non-array leaves and consider them as static. This
    results in module instances compatible with native JAX transformations
    (:func:`jax.jit`, :func:`jax.vmap`, :func:`jax.grad`, ...) out of the box.

    .. code-block:: python

        import optax

        @jax.jit
        def loss_fn(model, x, y):
            logits = jax.vmap(model)(x)
            loss = optax.softmax_cross_entropy(logits, y)

            return jax.numpy.mean(loss)

        grads = jax.grad(loss_fn)(model, x, y)

    However, JAX transformations are designed to work on pure functions. Some neural
    network layers, including batch normalization, are not pure as they hold a state
    which is updated as part of the layer's execution. In this case, using a
    functionally pure definition of the model is safer for training, but also convenient
    as some internal arrays do not require gradients.

    .. code-block:: python

        modef, params, others = model.pure(nn.Parameter)
        optimizer = optax.adamw(learning_rate=1e-3)
        opt_state = optimizer.init(params)

        @jax.jit
        def step(params, others, opt_state, x, y):  # gradient descent step
            def loss_fn(params):
                model = modef(params, others)
                logits = jax.vmap(model)(x)
                loss = optax.softmax_cross_entropy(logits, y)
                _, _, others = model.pure(nn.Parameter)

                return jax.numpy.mean(loss), others

            grads, others = jax.grad(loss_fn, has_aux=True)(params)
            updates, opt_state = optimizer.update(grads, opt_state, params)
            params = optax.apply_updates(params, updates)

            return params, others, opt_state

        for x, y in trainset:  # training loop
            params, others, opt_state = step(params, others, opt_state, x, y)

        model = modef(params, others)

    Arguments:
        kwargs: A name-value mapping.
    """

    def train(self, mode: bool = True):
        r"""Toggles between training and evaluation modes.

        This method is primarily useful for (sub)modules that behave differently at
        training and evaluation, such as :class:`inox.nn.dropout.TrainingDropout` and
        :class:`inox.nn.normalization.BatchNorm`.

        Arguments:
            mode: Whether to turn training mode on or off.

        Example:
            >>> model.train(False)  # turns off dropout
        """

        leaves = jtu.tree_leaves(vars(self), is_leaf=is_module)

        for leaf in leaves:
            if is_module(leaf):
                leaf.train(mode)

        if hasattr(self, 'training'):
            self.training = mode

    def pure(self, *roles: type) -> Tuple[ModuleDef, Dict[str, Array]]:
        r"""Splits the functional definition of the module from its state.

        The state is represented by a path-array mapping split into different
        collections depending on the role (:class:`Parameter`, :class:`Statistic`, ...)
        of arrays. One collection is returned for each provided role plus an additional
        collection for arrays that do not fit any role.

        See also:
            :class:`ModuleDef`

        Arguments:
            roles: A set of :class:`Variable` subclasses.

        Returns:
            The module definition and state collection(s).

        Examples:
            >>> modef, state = model.pure()
            >>> clone = modef(state)

            >>> modef, params, others = model.pure(nn.Parameter)
            >>> params, opt_state = optimizer.update(grads, opt_state, params)
            >>> model = modef(params, others)
        """

        assert all(issubclass(role, Variable) for role in roles)

        state = [{} for _ in roles]
        state.append({})

        def is_leaf(x):
            return any(isinstance(x, role) for role in roles)

        leaves = jtu.tree_leaves_with_path(tree=self, is_leaf=is_leaf)

        for path, leaf in leaves:
            key = jtu.keystr(path)

            for i, role in enumerate(roles):
                if isinstance(leaf, role):
                    state[i][key] = leaf.value
                    break
            else:
                state[-1][key] = leaf

        return ModuleDef(self), *state


class ModuleDef(Static):
    r"""Abstraction for the static definition of a module.

    See also:
        :meth:`Module.pure`

    Arguments:
        module: A module instance.
    """

    def __init__(self, module: Module):
        self.value = jtu.tree_structure(module)

    def __call__(self, *state: Dict[str, Array]) -> Module:
        r"""
        Arguments:
            state: A set of state collections.

        Returns:
            A new instance of the module.
        """

        state = {
            key: leaf
            for collection in state
            for key, leaf in collection.items()
        }

        missing = []

        def f(path, leaf):
            key = jtu.keystr(path)

            if key in state:
                leaf = state.pop(key)
            else:
                missing.append(key)

            return leaf

        leaves = [object()] * self.value.num_leaves
        module = jtu.tree_unflatten(self.value, leaves)
        module = jtu.tree_map_with_path(f=f, tree=module)

        if missing:
            keys = ', '.join(f'"{key}"' for key in missing)

            raise KeyError(f"Missing key(s) in state: {keys}.")

        if state:
            keys = ', '.join(f'"{key}"' for key in state)

            raise KeyError(f"Unexpected key(s) in state: {keys}.")

        return module


class Variable(PyArray):
    r"""Wrapper to indicate variable arrays.

    Wrapping arrays in :class:`Variable` subclasses allows to split the state of modules
    into sub-collections.

    Arguments:
        value: An array value.
    """

    pass


class Parameter(Variable):
    r"""Wrapper to indicate optimizable arrays.

    All arrays that require gradient updates in a :class:`Module` should be wrapped in a
    :class:`Parameter` instance.

    Arguments:
        value: An array value.
    """

    pass


class Statistic(Variable):
    r"""Wrapper to indicate statistic arrays.

    See also:
        :class:`inox.nn.normalization.BatchNorm`

    Arguments:
        value: An array value.
    """

    pass



r"""Extended utilities for random number generation"""

__all__ = [
    'PRNG',
    'set_rng',
    'get_rng',
]

import jax

from contextlib import contextmanager
from jax import Array
from typing import *

from .debug import same_trace
from .tree_util import Namespace


class PRNG(Namespace):
    r"""Creates a pseudo-random number generator (PRNG).

    This class is a thin wrapper around the :mod:`jax.random` module, and allows to
    generate new PRNG keys or sample from distributions without having to split keys
    with :func:`jax.random.split` by hand.

    Arguments:
        seed: An integer seed or PRNG key.
        kwargs: Keyword arguments passed to :func:`jax.random.PRNGKey`.

    Example:
        >>> rng = PRNG(42)
        >>> rng.split()  # generates a key
        Array([2465931498, 3679230171], dtype=uint32)
        >>> rng.split(3)  # generates a vector of 3 keys
        Array([[ 956272045, 3465119146],
               [1903583750,  988321301],
               [3226638877, 2833683589]], dtype=uint32)
        >>> rng.normal((5,))
        Array([ 0.5694761 , -1.4582146 ,  0.2309113 , -0.03029377,  0.11095619], dtype=float32)
    """

    def __init__(self, seed: Union[int, Array], **kwargs):
        if isinstance(seed, int):
            self.state = jax.random.PRNGKey(seed, **kwargs)
        else:
            self.state = seed

    def __getattr__(self, name: str) -> Any:
        attr = getattr(jax.random, name)

        if callable(attr):
            return lambda *args, **kwargs: attr(self.split(), *args, **kwargs)
        else:
            return attr

    def split(self, num: int = None) -> Array:
        r"""
        Arguments:
            num: The number of keys to generate.

        Returns:
            A new key if :py:`num=None` and a vector of keys otherwise.
        """

        if num is None:
            keys = jax.random.split(self.state, num=2)
        else:
            keys = jax.random.split(self.state, num=num + 1)

        assert same_trace(self.state, keys), "the PRNG was initialized and used within different JIT traces."

        if num is None:
            key, self.state = keys
        else:
            key, self.state = keys[:-1], keys[-1]

        return key


INOX_RNG: PRNG = None


@contextmanager
def set_rng(rng: PRNG):
    r"""Sets the PRNG within a context.

    See also:
        :class:`PRNG` and :func:`get_rng`

    Arguments:
        rng: A PRNG instance.

    Example:
        >>> with set_rng(PRNG(0)):
        >>> ... a = get_rng().split()
        >>> ... b = get_rng().normal((2, 3))
    """

    global INOX_RNG

    try:
        old, INOX_RNG = INOX_RNG, rng
        yield
    finally:
        INOX_RNG = old


def get_rng() -> PRNG:
    r"""Returns the context-bound PRNG.

    See also:
        :class:`PRNG` and :func:`set_rng`
    """

    global INOX_RNG
    assert INOX_RNG is not None, "no PRNG is set in this context."
    return INOX_RNG



r"""Normalization layers"""

__all__ = [
    'BatchNorm',
    'LayerNorm',
]

import jax
import jax.numpy as jnp

from jax import Array
from typing import *

from .module import Module, Statistic
from ..debug import same_trace


class BatchNorm(Module):
    r"""Creates a batch-normalization layer.

    .. math:: y = \frac{x - \mathbb{E}[x]}{\sqrt{\mathbb{V}[x] + \epsilon}}

    The mean and variance are calculated over the batch and spatial axes. During
    training, the layer keeps running estimates of the mean and variance, which are then
    used for normalization during evaluation. The update rule for a running average
    statistic :math:`\hat{s}` is

    .. math:: \hat{s} \gets \alpha \hat{s} + (1 - \alpha) s

    where :math:`s` is the statistic calculated for the current batch.

    References:
        | Accelerating Deep Network Training by Reducing Internal Covariate Shift (Ioffe et al., 2015)
        | https://arxiv.org/abs/1502.03167

    Arguments:
        channels: The number of channels :math:`C`.
        epsilon: A numerical stability term :math:`\epsilon`.
        momentum: The momentum :math:`\alpha \in [0, 1]` for the running estimates.
    """

    training: bool = True

    def __init__(
        self,
        channels: int,
        epsilon: float = 1e-05,
        momentum: float = 0.9,
    ):
        self.epsilon = epsilon
        self.momentum = momentum

        self.mean = Statistic(jnp.zeros(channels))
        self.var = Statistic(jnp.ones(channels))

    def __call__(self, x: Array) -> Array:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:`(N, *, C)`.

        Returns:
            The output tensor :math:`y`, with shape :math:`(N, *, C)`.
        """

        if self.training:
            y = x.reshape(-1, x.shape[-1])
            mean = jnp.mean(y, axis=0)
            var = jnp.var(y, axis=0)

            self.mean.value = self.ema(self.mean, jax.lax.stop_gradient(mean))
            self.var.value = self.ema(self.var, jax.lax.stop_gradient(var))
        else:
            mean = self.mean
            var = self.var

        return (x - mean) / jnp.sqrt(var + self.epsilon)

    def ema(self, x: Array, y: Array) -> Array:
        assert same_trace(x, y), "an unsafe side effect was detected. Ensure that the running statistic and input arrays have the same trace."

        return self.momentum * x + (1 - self.momentum) * y


class LayerNorm(Module):
    r"""Creates a layer-normalization layer.

    .. math:: y = \frac{x - \mathbb{E}[x]}{\sqrt{\mathbb{V}[x] + \epsilon}}

    References:
        | Layer Normalization (Ba et al., 2016)
        | https://arxiv.org/abs/1607.06450

    Arguments:
        axis: The axis(es) over which the mean and variance are calculated.
        epsilon: A numerical stability term :math:`\epsilon`.
    """

    def __init__(
        self,
        axis: Union[int, Sequence[int]] = -1,
        epsilon: float = 1e-05,
    ):
        self.axis = axis
        self.epsilon = epsilon

    def __call__(self, x: Array) -> Array:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:`(*, C)`.

        Returns:
            The output tensor :math:`y`, with shape :math:`(*, C)`.
        """

        mean = jnp.mean(x, axis=self.axis, keepdims=True)
        var = jnp.var(x, axis=self.axis, keepdims=True)

        return (x - mean) / jnp.sqrt(var + self.epsilon)



r"""Container modules"""

__all__ = [
    'Sequential',
]

import jax

from textwrap import indent
from typing import *

from .module import Module
from ..tree_util import tree_repr


class Sequential(Module):
    r"""Creates a composition of layers.

    .. math:: y = f_n \circ \dots \circ f_2 \circ f_1(x)

    Arguments:
        layers: A sequence of layers :math:`f_1, f_2, \dots, f_n`.
    """

    def __init__(self, *layers: Module):
        self.layers = layers

    def __call__(self, x: Any) -> Any:
        r"""
        Arguments:
            x: The input :math:`x`.

        Returns:
            The output :math:`y`.
        """

        for layer in self.layers:
            x = layer(x)
        return x

    def tree_repr(self, **kwargs) -> str:
        lines = (tree_repr(layer, **kwargs) for layer in self.layers)
        lines = ',\n'.join(lines)

        if lines:
            lines = '\n' + indent(lines, '  ') + '\n'

        return f'{self.__class__.__name__}({lines})'



r"""Extended utilities for debugging"""

__all__ = [
    'same_trace',
]

import jax

from jax import Array
from jax.core import Tracer
from typing import *


def same_trace(a: Array, b: Array, ignore_primal: bool = False) -> bool:
    r"""Checks whether two arrays have the same trace source.

    Arguments:
        a: The first array.
        b: The second array.
        ignore_primal: Whether to ignore primal traces (:func:`jax.grad`).

    Example:
        >>> x, y = jax.numpy.zeros(2)
        >>> same_trace(x, y)
        True
        >>> jax.jit(lambda x, y: same_trace(x, y))(x, y)
        Array(True, dtype=bool)
        >>> jax.jit(lambda x: same_trace(x, y))(x)
        Array(False, dtype=bool)
    """

    if hasattr(a, '__jax_array__'):
        a = a.__jax_array__()

    if hasattr(b, '__jax_array__'):
        b = b.__jax_array__()

    if ignore_primal:
        while hasattr(a, 'primal'):
            a = a.primal

        while hasattr(b, 'primal'):
            b = b.primal

    if isinstance(a, Tracer) and isinstance(b, Tracer):
        return a._trace.main == b._trace.main
    elif isinstance(a, Tracer) or isinstance(b, Tracer):
        return False
    else:
        return True



r"""Recurrent layers"""

__all__ = [
    'Cell',
    'Recurrent',
    'BRCell',
    'MGUCell',
    'GRUCell',
    'LSTMCell',
]

import jax
import jax.numpy as jnp

from jax import Array
from typing import *

from .linear import Linear
from .module import Module


class Cell(Module):
    r"""Abstract cell class.

    A cell defines a recurrence function :math:`f` of the form

    .. math:: (h_i, y_i) = f(h_{i-1}, x_i)

    and an initial hidden state :math:`h_0`.

    Warning:
        The recurrence function :math:`f` should have no side effects.
    """

    def __call__(self, h: Any, x: Any) -> Tuple[Any, Any]:
        r"""
        Arguments:
            h: The previous hidden state :math:`h_{i-1}`.
            x: The input :math:`x_i`.

        Returns:
            The hidden state and output :math:`(h_i, y_i)`.
        """

        raise NotImplementedError()

    def init(self) -> Any:
        r"""
        Returns:
            The initial hidden state :math:`h_0`.
        """

        raise NotImplementedError()


class Recurrent(Module):
    r"""Creates a recurrent layer.

    Arguments:
        cell: A recurrent cell.
        reverse: Whether to apply the recurrence in reverse or not.
    """

    def __init__(
        self,
        cell: Cell,
        reverse: bool = False,
    ):
        self.cell = cell
        self.reverse = reverse

    def __call__(self, xs: Any) -> Any:
        r"""
        Arguments:
            xs: A sequence of inputs :math:`x_i`, stacked on the leading axis.
                When inputs are vectors, :py:`xs` has shape :math:`(L, C)`.

        Returns:
            A sequence of outputs :math:`y_i`, stacked on the leading axis. When outputs
            are vectors, :py:`ys` has shape :math:`(L, C')`.
        """

        _, ys = jax.lax.scan(
            f=self.cell,
            init=self.cell.init(),
            xs=xs,
            reverse=self.reverse,
        )

        return ys


class GRUCell(Cell):
    r"""Creates a gated recurrent unit (GRU) cell.

    References:
        | Learning Phrase Representations using RNN Encoder-Decoder for Statistical Machine Translation (Cho et al., 2014)
        | https://arxiv.org/abs/1406.1078

    Arguments:
        key: A PRNG key for initialization.
        in_features: The number of input features :math:`C`.
        hid_features: The number of hidden features :math:`H`.
        bias: Whether the cell learns additive biases or not.
    """

    def __init__(
        self,
        key: Array,
        in_features: int,
        hid_features: int,
        bias: bool = True,
    ):
        keys = jax.random.split(key, 2)

        self.lin_h = Linear(keys[0], hid_features, 3 * hid_features, bias)
        self.lin_x = Linear(keys[1], in_features, 3 * hid_features, bias)

        self.in_features = in_features
        self.hid_features = hid_features

    def __call__(self, h: Array, x: Array) -> Tuple[Array, Array]:
        r"""
        Arguments:
            h: The previous hidden state :math:`h_{i-1}`, with shape :math:`(*, H)`.
            x: The input vector :math:`x_i`, with shape :math:`(*, C)`.

        Returns:
            The hidden state :math:`(h_i, h_i)`.
        """

        rh, zh, gh = jnp.split(self.lin_h(h), 3, axis=-1)
        rx, zx, gx = jnp.split(self.lin_x(x), 3, axis=-1)

        r = jax.nn.sigmoid(rx + rh)
        z = jax.nn.sigmoid(zx + zh)
        g = jax.nn.tanh(gx + r * gh)

        h = (1 - z) * g + z * h

        return h, h

    def init(self) -> Array:
        r"""
        Returns:
            The initial hidden state :math:`h_0 = 0`, with shape :math:`(H)`.
        """

        return jnp.zeros(self.hid_features)


class BRCell(Module):
    r"""Creates a bistable recurrent cell (BRC).

    References:
        | A bio-inspired bistable recurrent cell allows for long-lasting memory (Vecoven et al., 2021)
        | https://arxiv.org/abs/2006.05252

    Arguments:
        key: A PRNG key for initialization.
        in_features: The number of input features :math:`C`.
        hid_features: The number of hidden features :math:`H`.
        bias: Whether the cell learns additive biases or not.
        modulated: Whether to use neuromodulation or not.
    """

    def __init__(
        self,
        key: Array,
        in_features: int,
        hid_features: int,
        bias: bool = True,
        modulated: bool = True,
    ):
        keys = jax.random.split(key, 3)

        self.modulated = modulated

        self.lin_x = Linear(keys[0], in_features, 3 * hid_features, bias)

        if self.modulated:
            self.lin_h = Linear(keys[1], hid_features, 2 * hid_features, bias)
        else:
            self.wa = jax.random.normal(keys[1], (hid_features,))
            self.wc = jax.random.normal(keys[2], (hid_features,))

        self.in_features = in_features
        self.hid_features = hid_features

    def __call__(self, h: Array, x: Array) -> Tuple[Array, Array]:
        r"""
        Arguments:
            h: The previous hidden state :math:`h_{i-1}`, with shape :math:`(*, H)`.
            x: The input vector :math:`x_i`, with shape :math:`(*, C)`.

        Returns:
            The hidden state :math:`(h_i, h_i)`.
        """

        if self.modulated:
            ah, ch = jnp.split(self.lin_h(h), 2, axis=-1)
        else:
            ah = self.wa * h
            ch = self.wc * h

        ax, cx, gx = jnp.split(self.lin_x(x), 3, axis=-1)

        a = 1.0 + jax.nn.tanh(ax + ah)
        c = jax.nn.sigmoid(cx + ch)
        g = jax.nn.tanh(gx + a * h)

        h = (1 - c) * g + c * h

        return h, h

    def init(self) -> Array:
        r"""
        Returns:
            The initial hidden state :math:`h_0 = 0`, with shape :math:`(H)`.
        """

        return jnp.zeros(self.hid_features)


class MGUCell(Module):
    r"""Creates a minimal gated unit (MGU) cell.

    References:
        | Minimal Gated Unit for Recurrent Neural Networks (Zhou et al., 2016)
        | https://arxiv.org/pdf/1603.09420

    Arguments:
        key: A PRNG key for initialization.
        in_features: The number of input features :math:`C`.
        hid_features: The number of hidden features :math:`H`.
        bias: Whether the cell learns additive biases or not.
    """

    def __init__(
        self,
        key: Array,
        in_features: int,
        hid_features: int,
        bias: bool = True,
    ):
        keys = jax.random.split(key, 2)

        self.lin_fh = Linear(keys[0], hid_features, 1 * hid_features, bias)
        self.lin_hh = Linear(keys[0], hid_features, 1 * hid_features, bias)
        self.lin_x = Linear(keys[1], in_features, 2 * hid_features, bias)

        self.in_features = in_features
        self.hid_features = hid_features

    def __call__(self, h: Array, x: Array) -> Tuple[Array, Array]:
        r"""
        Arguments:
            h: The previous hidden state :math:`h_{i-1}`, with shape :math:`(*, H)`.
            x: The input vector :math:`x_i`, with shape :math:`(*, C)`.

        Returns:
            The hidden state :math:`(h_i, h_i)`.
        """

        fh = self.lin_fh(h)
        fx, gx = jnp.split(self.lin_x(x), 2, axis=-1)
        f = jax.nn.sigmoid(fx + fh)
        gh = self.lin_hh(f * h)
        g = jax.nn.tanh(gx + gh)
        h = (1 - f) * g + f * h

        return h, h

    def init(self) -> Array:
        r"""
        Returns:
            The initial hidden state :math:`h_0 = 0`, with shape :math:`(H)`.
        """

        return jnp.zeros(self.hid_features)


class LSTMCell(Cell):
    r"""Creates a long short-term memory (LSTM) cell.

    References:
        | Long Short-Term Memory (Hochreiter et al., 1997)
        | https://ieeexplore.ieee.org/abstract/document/6795963

    Arguments:
        key: A PRNG key for initialization.
        in_features: The number of input features :math:`C`.
        hid_features: The number of hidden features :math:`H`.
        bias: Whether the cell learns additive biases or not.
    """

    def __init__(
        self,
        key: Array,
        in_features: int,
        hid_features: int,
        bias: bool = True,
    ):
        keys = jax.random.split(key, 2)

        self.lin_h = Linear(keys[0], hid_features, 4 * hid_features, bias)
        self.lin_x = Linear(keys[1], in_features, 4 * hid_features, bias)

        self.in_features = in_features
        self.hid_features = hid_features

    def __call__(
        self,
        hc: Tuple[Array, Array],
        x: Array,
    ) -> Tuple[Tuple[Array, Array], Array]:
        r"""
        Arguments:
            hc: The previous hidden and cell states :math:`(h_{i-1}, c_{i-1})`,
                each with shape :math:`(*, H)`.
            x: The input vector :math:`x_i`, with shape :math:`(*, C)`.

        Returns:
            The hidden and cell states :math:`((h_i, c_i), h_i)`.
        """

        h, c = hc
        i, f, g, o = jnp.split(self.lin_h(h) + self.lin_x(x), 4, axis=-1)

        i = jax.nn.sigmoid(i)
        f = jax.nn.sigmoid(f)
        g = jax.nn.tanh(g)
        o = jax.nn.sigmoid(o)

        c = f * c + i * g
        h = o * jax.nn.tanh(c)

        return (h, c), h

    def init(self) -> Tuple[Array, Array]:
        r"""
        Returns:
            The initial hidden and cell states :math:`h_0 = c_0 = 0`,
            each with shape :math:`(H)`.
        """

        return jnp.zeros(self.hid_features), jnp.zeros(self.hid_features)



r"""Tests for the inox.debug module."""

import jax
import pytest

from functools import partial
from inox.debug import *
from inox.tree_util import PyArray


@pytest.mark.parametrize('pyarray', [False, True])
def test_same_trace(pyarray: bool):
    x = jax.numpy.zeros(1)
    y = jax.numpy.ones(1)

    if pyarray:
        y = PyArray(y)

    assert same_trace(x, y)
    assert jax.jit(same_trace)(x, y)
    assert not jax.jit(partial(same_trace, x))(y)



r"""Upstream patches"""

__all__ = []

from jax._src import core
from jax._src.lax import lax


# core.Primitive.bind comply with 'valid_jaxtype'
bind_old = core.Primitive.bind

def bind_new(self, *args, **kwargs):
    args = [
        x.__jax_array__() if hasattr(x, '__jax_array__') else x
        for x in args
    ]

    return bind_old(self, *args, **kwargs)

core.Primitive.bind = bind_new


# lax.asarray comply with 'valid_jaxtype'
asarray_old = lax.asarray

def asarray_new(x):
    if hasattr(x, '__jax_array__'):
        x = x.__jax_array__()

    return asarray_old(x)

lax.asarray = asarray_new



r"""Padding layers"""

__all__ = [
    'Pad',
]

import jax

from jax import Array
from typing import *

from .module import Module
from ..numpy import vectorize


class Pad(Module):
    r"""Creates a spatial padding layer.

    This module is a thin wrapper around :func:`jax.numpy.pad`.

    Arguments:
        padding: The padding applied to each end of each spatial axis.
        mode: The padding mode in :py:`{'constant', 'edge', 'reflect', 'wrap'}`.
        value: The padding value if :py:`mode='constant`.
    """

    def __init__(
        self,
        padding: Sequence[Tuple[int, int]],
        mode: str = 'constant',
        value: float = 0.0,
    ):
        self.padding = padding
        self.mode = mode
        self.value = value

    def __call__(self, x: Array) -> Array:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:`(*, H_1, \dots, H_n, C)`.

        Returns:
            The output tensor :math:`y`, with shape :math:`(*, H_1', \dots, H_n', C)`,
            such that

            .. math:: H_i' = H_i + p_i

            where :math:`p_i` is the total padding of the :math:`i`-th spatial axis.
        """

        return vectorize(jax.numpy.pad, ndims=self.ndim)(
            x,
            pad_width=(*self.padding, (0, 0)),
            mode=self.mode,
            constant_values=self.value,
        )

    @property
    def ndim(self) -> int:
        return len(self.padding) + 1



r"""Attention layers"""

__all__ = [
    'MultiheadAttention',
]

import jax
import math

from einops import rearrange
from jax import Array
from typing import *

from .linear import Linear
from .module import Module


def attention(
    q: Array,
    k: Array,
    v: Array,
    mask: Array = None,
) -> Array:
    r"""Computes the scaled dot-product attention.

    Arguments:
        q: The query tensor :math:`Q`, with shape :math:`(*, S, C)`.
        k: The key tensor :math:`K`, with shape :math:`(*, T, C)`.
        v: The value tensor :math:`V`, with shape :math:`(*, T, C')`.
        mask: A boolean attention mask, with shape :math:`(*, S, T)`.
            A :py:`False` value indicates that the corresponding attention weight
            is set to :math:`-\infty`.

    Returns:
        The output vector :math:`y`, with shape :math:`(*, S, C')`.
    """

    C = q.shape[-1]

    weight = jax.numpy.einsum('...ik,...jk->...ij', q, k)
    weight = weight / math.sqrt(C)

    if mask is not None:
        weight = jax.numpy.where(mask, weight, -1e9)

    attn = jax.nn.softmax(weight, axis=-1)

    return jax.numpy.einsum('...ij,...jk->...ik', attn, v)


class MultiheadAttention(Module):
    r"""Creates a multihead attention layer.

    .. math:: Y = \sum_i
        \mathrm{attention}(X_q W_q^i + b_q^i, X_k W_k^i + b_k^i, X_v W_v^i) W_y^i

    where

    .. math:: \mathrm{attention}(Q, K, V) =
        \mathrm{softmax}\left( \frac{Q K^T}{\sqrt{H}} \right) V

    denotes the scaled dot-product attention.

    References:
        | Attention Is All You Need (Vaswani et al., 2023)
        | https://arxiv.org/abs/1706.03762

    Arguments:
        key: A PRNG key for initialization.
        heads: The number of attention heads.
        in_features: The number of input features :math:`C`.
        hid_features: The number of hidden features :math:`H` per head.
        out_features: The number of output features :math:`C'`.
            If :py:`None`, :math:`C' = C`.
        bias: Whether the layer learns additive biases :math:`(b_q, b_k)` or not.
        causal: Whether the attention mask is causal or not. If :py:`True`, the
            :math:`i`-th query is only allowed to attend the :math:`j`-th key if
            :math:`j - i \leq T - S`.
        dropout: The dropout rate on attention weights.
    """

    def __init__(
        self,
        key: Array,
        heads: int,
        in_features: int,
        hid_features: int,
        out_features: int = None,
        bias: bool = True,
        causal: bool = False,
        dropout: float = 0.0,
    ):
        keys = jax.random.split(key, 4)

        if out_features is None:
            out_features = in_features

        self.lin_q = Linear(keys[0], in_features, hid_features * heads, bias=bias)
        self.lin_k = Linear(keys[1], in_features, hid_features * heads, bias=bias)
        self.lin_v = Linear(keys[2], in_features, hid_features * heads, bias=False)
        self.lin_y = Linear(keys[3], hid_features * heads, out_features, bias=False)

        self.heads = heads
        self.causal = causal
        self.dropout = dropout

    def __call__(
        self,
        xq: Array,
        xk: Array = None,
        xv: Array = None,
        mask: Array = None,
        key: Array = None,
    ) -> Array:
        r"""
        Arguments:
            xq: The query tensor :math:`X_q`, with shape :math:`(*, S, C)`.
            xk: The key tensor :math:`X_k`, with shape :math:`(*, T, C)`.
                If :py:`None`, :math:`X_k = X_q`.
            xv: The value tensor :math:`X_v`, with shape :math:`(*, T, C)`.
                If :py:`None`, :math:`X_v = X_k`.
            mask: A boolean attention mask, with shape :math:`(*, S, T)`.
                A :py:`False` value indicates that the corresponding attention weight
                is set to :math:`-\infty`.
            key: A PRNG key. If :py:`None`, dropout is not applied.

        Returns:
            The output tensor :math:`Y`, with shape :math:`(*, S, C')`.
        """

        if xk is None:
            xk = xq

        if xv is None:
            xv = xk

        S, T = xq.shape[-2], xk.shape[-2]

        # Project
        q = self.lin_q(xq)
        k = self.lin_k(xk)
        v = self.lin_v(xv)

        q, k, v = [
            rearrange(x, '... L (N H) -> ... N L H', H=self.heads)
            for x in (q, k, v)
        ]

        # Mask
        if self.causal:
            if mask is None:
                mask = jax.numpy.ones((S, T), dtype=bool)

            mask = jax.numpy.tril(mask, T - S)

        if key is not None:
            shape = jax.numpy.broadcast_shapes(
                (*q.shape[:-2], S, 1),
                (*k.shape[:-2], 1, T),
                (S, T) if mask is None else mask.shape,
            )

            keep = jax.random.bernoulli(key, p=1 - self.dropout, shape=shape)

            if mask is None:
                mask = keep
            else:
                mask = jax.numpy.logical_and(mask, keep)

        # Attention
        y = attention(q, k, v, mask)
        y = rearrange(y, '... N L H -> ... L (N H)')
        y = self.lin_y(y)

        return y



r"""Linear layers"""

__all__ = [
    'Linear',
    'Conv',
    'ConvTransposed',
]

import jax
import math

from jax import Array
from typing import *

from .module import Module, Parameter
from ..numpy import flatten, unflatten


class Linear(Module):
    r"""Creates a linear layer.

    .. math:: y = W x + b

    Arguments:
        key: A PRNG key for initialization.
        in_features: The number of input features :math:`C`.
        out_features: The number of output features :math:`C'`.
        bias: Whether the layer learns an additive bias :math:`b` or not.
    """

    def __init__(
        self,
        key: Array,
        in_features: int,
        out_features: int,
        bias: bool = True,
    ):
        keys = jax.random.split(key, 2)
        lim = 1 / math.sqrt(in_features)

        self.weight = Parameter(
            jax.random.uniform(
                key=keys[0],
                shape=(in_features, out_features),
                minval=-lim,
                maxval=lim,
            )
        )

        if bias:
            self.bias = Parameter(
                jax.random.uniform(
                    key=keys[1],
                    shape=(out_features,),
                    minval=-lim,
                    maxval=lim,
                )
            )
        else:
            self.bias = None

    def __call__(self, x: Array) -> Array:
        r"""
        Arguments:
            x: The input vector :math:`x`, with shape :math:`(*, C)`.

        Returns:
            The output vector :math:`y`, with shape :math:`(*, C')`.
        """

        if self.bias is None:
            return x @ self.weight
        else:
            return x @ self.weight + self.bias


class Conv(Module):
    r"""Creates a convolution layer.

    .. math:: y = W * x + b

    References:
        | A guide to convolution arithmetic for deep learning (Dumoulin et al., 2016)
        | https://arxiv.org/abs/1603.07285

    Arguments:
        key: A PRNG key for initialization.
        in_channels: The number of input channels :math:`C`.
        out_channels: The number of output channels :math:`C'`.
        kernel_size: The size of the kernel :math:`W` in each spatial axis.
        bias: Whether the layer learns an additive bias :math:`b` or not.
        stride: The stride coefficient in each spatial axis.
        dilation: The dilation coefficient in each spatial axis.
        padding: The padding applied to each end of each spatial axis.
        groups: The number of channel groups :math:`G`.
            Both :math:`C` and :math:`C'` must be divisible by :math:`G`.
    """

    def __init__(
        self,
        key: Array,
        in_channels: int,
        out_channels: int,
        kernel_size: Sequence[int],
        bias: bool = True,
        stride: Union[int, Sequence[int]] = 1,
        dilation: Union[int, Sequence[int]] = 1,
        padding: Union[int, Sequence[Tuple[int, int]]] = 0,
        groups: int = 1,
    ):
        in_channels = in_channels // groups

        if isinstance(stride, int):
            stride = [stride] * len(kernel_size)

        if isinstance(dilation, int):
            dilation = [dilation] * len(kernel_size)

        if isinstance(padding, int):
            padding = [(padding, padding)] * len(kernel_size)

        keys = jax.random.split(key, 2)
        lim = 1 / math.sqrt(math.prod(kernel_size) * in_channels)

        self.kernel = Parameter(
            jax.random.uniform(
                key=keys[0],
                shape=(*kernel_size, in_channels, out_channels),
                minval=-lim,
                maxval=lim,
            )
        )

        if bias:
            self.bias = Parameter(
                jax.random.uniform(
                    key=keys[1],
                    shape=(out_channels,),
                    minval=-lim,
                    maxval=lim,
                )
            )
        else:
            self.bias = None

        self.kernel_size = kernel_size
        self.stride = stride
        self.dilation = dilation
        self.padding = padding
        self.groups = groups

    def __call__(self, x: Array) -> Array:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:`(*, H_1, \dots, H_n, C)`.

        Returns:
            The output tensor :math:`y`, with shape :math:`(*, H_1', \dots, H_n', C')`,
            such that

            .. math:: H_i' =
                \left\lfloor \frac{H_i - d_i \times (k_i - 1) + p_i}{s_i} + 1 \right\rfloor

            where :math:`k_i`, :math:`s_i`, :math:`d_i` and :math:`p_i` are respectively
            the kernel size, the stride coefficient, the dilation coefficient and the
            total padding of the :math:`i`-th spatial axis.
        """

        batch = x.shape[: -self.ndim]

        x = flatten(x, 0, -self.ndim)
        x = jax.lax.conv_general_dilated(
            lhs=x,
            rhs=self.kernel,
            dimension_numbers=self.dimensions,
            window_strides=self.stride,
            rhs_dilation=self.dilation,
            padding=self.padding,
            feature_group_count=self.groups,
        )
        x = unflatten(x, 0, batch)

        if self.bias is None:
            return x
        else:
            return x + self.bias

    @property
    def ndim(self) -> int:
        return len(self.kernel_size) + 1

    @property
    def dimensions(self) -> jax.lax.ConvDimensionNumbers:
        return jax.lax.ConvDimensionNumbers(
            (0, self.ndim, *range(1, self.ndim)),
            (self.ndim, self.ndim - 1, *range(0, self.ndim - 1)),
            (0, self.ndim, *range(1, self.ndim)),
        )


class ConvTransposed(Conv):
    r"""Creates a transposed convolution layer.

    This layer can be seen as the gradient of :class:`Conv` with respect to its input.
    It is also known as a "deconvolution", altough it does not actually compute the
    inverse of a convolution.

    References:
        | A guide to convolution arithmetic for deep learning (Dumoulin et al., 2016)
        | https://arxiv.org/abs/1603.07285

    Arguments:
        key: A PRNG key for initialization.
        in_channels: The number of input channels :math:`C`.
        out_channels: The number of output channels  :math:`C'`.
        kernel_size: The size of the kernel :math:`W` in each spatial axis.
        bias: Whether the layer learns an additive bias :math:`b` or not.
        stride: The stride coefficient in each spatial axis.
        dilation: The dilation coefficient in each spatial axis.
        padding: The padding applied to each end of each spatial axis.
        groups: The number of channel groups :math:`G`.
            Both :math:`C` and :math:`C'` must be divisible by :math:`G`.
    """

    def __call__(self, x: Array) -> Array:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:`(*, H_1, \dots, H_n, C)`.

        Returns:
            The output tensor :math:`y`, with shape :math:`(*, H_1', \dots, H_n', C')`,
            such that

            .. math:: H_i' = (H_i - 1) \times s_i + d_i \times (k_i - 1) - p_i + 1

            where :math:`k_i`, :math:`s_i`, :math:`d_i` and :math:`p_i` are respectively
            the kernel size, the stride coefficient, the dilation coefficient and the
            total padding of the :math:`i`-th spatial axis.
        """

        batch = x.shape[: -self.ndim]

        x = flatten(x, 0, -self.ndim)
        x = jax.lax.conv_general_dilated(
            lhs=x,
            rhs=self.kernel,
            dimension_numbers=self.dimensions,
            window_strides=[1] * (self.ndim - 1),
            padding=self.transposed_padding,
            lhs_dilation=self.stride,
            rhs_dilation=self.dilation,
            feature_group_count=self.groups,
        )
        x = unflatten(x, 0, batch)

        if self.bias is None:
            return x
        else:
            return x + self.bias

    @property
    def transposed_padding(self) -> Sequence[Tuple[int, int]]:
        return [
            (d * (k - 1) - p[0], d * (k - 1) - p[1])
            for k, d, p in zip(
                self.kernel_size,
                self.dilation,
                self.padding,
            )
        ]



r"""Activation functions"""

__all__ = [
    'Identity',
    'Tanh',
    'Sigmoid',
    'SiLU',
    'Softplus',
    'Softmax',
    'ReLU',
    'LeakyReLU',
    'ELU',
    'CELU',
    'GELU',
    'SELU',
    'SiLU',
]

import jax

from jax import Array
from typing import *

from .module import Module


class Activation(Module):
    r"""Abstract activation class."""

    def __init__(self):
        pass


class Identity(Activation):
    r"""Creates an identity activation function.

    .. math:: y = x
    """

    def __call__(self, x: Array) -> Array:
        return x


class Tanh(Activation):
    r"""Creates an identity activation function.

    .. math:: y = \tanh(x)
    """

    def __call__(self, x: Array) -> Array:
        return jax.numpy.tanh(x)


class Sigmoid(Activation):
    r"""Creates a sigmoid activation function.

    .. math:: y = \sigma(x) = \frac{1}{1 + \exp(-x)}
    """

    def __call__(self, x: Array) -> Array:
        return jax.nn.sigmoid(x)


class Softplus(Activation):
    r"""Creates a softplus activation function.

    .. math:: y = \log(1 + \exp(x))
    """

    def __call__(self, x: Array) -> Array:
        return jax.nn.softplus(x)


class Softmax(Activation):
    r"""Creates a softmax activation function.

    .. math:: y_i = \frac{\exp(x_i)}{\sum_j \exp(x_j)}

    Arguments:
        axis: The axis(es) over which the sum is performed.
    """

    def __init__(self, axis: Union[int, Sequence[int]] = -1):
        self.axis = axis

    def __call__(self, x: Array) -> Array:
        return jax.nn.softmax(x, axis=self.axis)

    def tree_repr(self, **kwargs) -> str:
        return f'{self.__class__.__name__}(axis={self.axis})'


class ReLU(Activation):
    r"""Creates a rectified linear unit (ReLU) activation function.

    .. math:: y = \max(x, 0)
    """

    def __call__(self, x: Array) -> Array:
        return jax.nn.relu(x)


class LeakyReLU(Activation):
    r"""Creates a leaky-ReLU activation function.

    .. math:: y = \begin{cases}
            \alpha x & \text{if } x \leq 0 \\
            x & \text{otherwise}
        \end{cases}

    Arguments:
        alpha: The negative slope :math:`\alpha`.
    """

    def __init__(self, alpha: float = 0.01):
        self.alpha = alpha

    def __call__(self, x: Array) -> Array:
        return jax.nn.leaky_relu(x, self.alpha)

    def tree_repr(self, **kwargs) -> str:
        return f'{self.__class__.__name__}(alpha={self.alpha})'


class ELU(Activation):
    r"""Creates an exponential linear unit (ELU) activation function.

    .. math:: y = \begin{cases}
            \alpha (\exp(x) - 1) & \text{if } x \leq 0 \\
            x & \text{otherwise}
        \end{cases}

    References:
        | Fast and Accurate Deep Network Learning by Exponential Linear Units (Clevert et al., 2015)
        | https://arxiv.org/abs/1511.07289

    Arguments:
        alpha: The coefficient :math:`\alpha`.
    """

    def __init__(self, alpha: float = 1.0):
        self.alpha = alpha

    def __call__(self, x: Array) -> Array:
        return jax.nn.elu(x)

    def tree_repr(self, **kwargs) -> str:
        return f'{self.__class__.__name__}(alpha={self.alpha})'


class CELU(ELU):
    r"""Creates a continuously-differentiable ELU (CELU) activation function.

    .. math:: y = \max(x, 0) + \alpha \min(0, \exp(x / \alpha) - 1)

    References:
        | Continuously Differentiable Exponential Linear Units (Barron, 2017)
        | https://arxiv.org/abs/1704.07483

    Arguments:
        alpha: The coefficient :math:`\alpha`.
    """

    def __call__(self, x: Array) -> Array:
        return jax.nn.celu(x)


class GELU(Activation):
    r"""Creates a Gaussian error linear unit (GELU) activation function.

    .. math:: y = \frac{x}{2}
        \left(1 + \mathrm{erf}\left(\frac{x}{\sqrt{2}}\right)\right)

    When :py:`approximate=True`, it is approximated as

    .. math:: y = \frac{x}{2}
        \left(1 + \tanh\left(\sqrt{\frac{2}{\pi}}(x + 0.044715 x^3)\right)\right)

    References:
        | Gaussian Error Linear Units (Hendrycks et al., 2017)
        | https://arxiv.org/abs/1606.08415v4

    Arguments:
        approximate: Whether to use the approximate or exact formulation.
    """

    def __init__(self, approximate: bool = True):
        self.approximate = approximate

    def __call__(self, x: Array) -> Array:
        return jax.nn.gelu(x, self.approximate)

    def tree_repr(self, **kwargs) -> str:
        return f'{self.__class__.__name__}(approximate={self.approximate})'


class SELU(Activation):
    r"""Creates a self-normalizing ELU (SELU) activation function.

    .. math:: y = \lambda \begin{cases}
            \alpha (\exp(x) - 1) & \text{if } x \leq 0 \\
            x & \text{otherwise}
        \end{cases}

    where :math:`\lambda \approx 1.0507` and :math:`\alpha \approx 1.6732`.

    References:
        | Self-Normalizing Neural Networks (Klambauer et al., 2017)
        | https://arxiv.org/abs/1706.02515
    """

    def __call__(self, x: Array) -> Array:
        return jax.nn.selu(x)


class SiLU(Activation):
    r"""Creates a sigmoid linear unit (SiLU) activation function.

    .. math:: y = x \sigma(x)

    References:
        | Gaussian Error Linear Units (Hendrycks et al., 2017)
        | https://arxiv.org/abs/1606.08415v4

        | Sigmoid-Weighted Linear Units for Neural Network Function Approximation in Reinforcement Learning (Elfwing et al., 2017)
        | https://arxiv.org/abs/1702.03118
    """

    def __call__(self, x: Array) -> Array:
        return jax.nn.silu(x)



r"""Neural networks, layers and modules"""

from .activation import *
from .attention import *
from .container import *
from .dropout import *
from .einops import *
from .linear import *
from .module import *
from .normalization import *
from .padding import *
from .pooling import *
from .recurrent import *
from .ssm import *



r"""Tests for the inox.nn module."""

import jax
import pickle
import pytest

from jax import Array
from typing import *

from inox.nn import *


def test_Module():
    module = Module(
        a=jax.numpy.zeros(()),
        b=Parameter(jax.numpy.ones(1)),
        c=[True, Statistic(jax.numpy.arange(3))],
        d='four',
    )

    # Flatten
    leaves, treedef = jax.tree_util.tree_flatten(module)

    assert all(isinstance(leaf, Array) for leaf in leaves)
    assert isinstance(treedef, Hashable)

    jax.tree_util.tree_unflatten(treedef, leaves)

    # Pure
    modef, state = module.pure()

    print(state)

    assert all(isinstance(leaf, Array) for leaf in state.values())
    assert isinstance(modef, Hashable)

    modef(state)

    # Pure split
    modef, params, others = module.pure(Parameter)

    assert all(isinstance(leaf, Array) for leaf in params.values())
    assert all(isinstance(leaf, Array) for leaf in others.values())
    assert isinstance(modef, Hashable)

    modef(params, others)

    with pytest.raises(KeyError):
        modef(params)

    with pytest.raises(KeyError):
        modef(params, others, {'e': None})

    # Pickle
    data = pickle.dumps(module)
    pickle.loads(data)

    # Print
    assert repr(module)


def test_MLP():
    key = jax.random.key(0)
    x = jax.random.normal(key, (1024, 3))
    y = jax.numpy.linalg.norm(x, axis=-1, keepdims=True)

    class MLP(Module):
        def __init__(self, key):
            keys = jax.random.split(key)

            self.l1 = Linear(keys[0], 3, 64)
            self.l2 = Linear(keys[1], 64, 1)
            self.relu = ReLU()

        def __call__(self, x):
            return self.l2(self.relu(self.l1(x)))

    # __init__
    model = MLP(key)

    # __call__
    z = model(x[0])
    z = model(x)
    z = jax.vmap(model)(x)

    # JIT
    @jax.jit
    def loss(model):
        return jax.numpy.mean((model(x) - y) ** 2)

    l = loss(model)

    # Gradients
    grads = jax.grad(loss)(model)



r"""Extended utilities for tree-like data structures"""

__all__ = [
    'PyArray',
    'Namespace',
    'Static',
    'Auto',
    'tree_repr',
]

import jax
import jax._src.tree_util as jtu
import numpy as np

from jax import Array
from textwrap import indent
from typing import *
from warnings import warn


PyTree = TypeVar('PyTree', bound=Any)
PyTreeDef = TypeVar('PyTreeDef')


def is_array(x: Any) -> bool:
    return isinstance(x, np.ndarray) or isinstance(x, Array)


def is_static(x: Any) -> bool:
    return isinstance(x, Static)


def is_auto(x: Any) -> bool:
    return isinstance(x, Auto)


class PyTreeMeta(type):
    r"""PyTree meta-class."""

    def __new__(cls, *args, **kwargs) -> type:
        cls = super().__new__(cls, *args, **kwargs)

        if hasattr(cls, 'tree_flatten_with_keys'):
            jtu.register_pytree_with_keys_class(cls)
        else:
            jtu.register_pytree_node_class(cls)

        return cls


class PyArray(metaclass=PyTreeMeta):
    r"""Wraps an array as a PyTree.

    Subclassing :class:`PyArray` allows to associate metadata (name, role, ...) to
    arrays, while preserving the array interface (`.shape`, `.dtype`, ...) and
    supporting :mod:`jax.numpy` operations.

    Arguments:
        value: An array value.

    Example:
        >>> x = jax.numpy.arange(5)
        >>> x = PyArray(x); x
        int32[5]
        >>> jax.numpy.mean(x)
        Array(2., dtype=float32)
    """

    value: Array = None

    def __init__(self, value: Array):
        self.value = value

    def __array__(self, dtype=None):
        return self.value.__array__(dtype)

    def __array_module__(self, types):
        return self.value.__array_module__(types)

    def __jax_array__(self) -> Array:
        if hasattr(self.value, '__jax_array__'):
            return self.value.__jax_array__()
        else:
            return self.value

    def __getattr__(self, attr: str) -> Any: return getattr(self.value, attr)
    def __getitem__(self, key: Hashable): return self.value[key]
    def __len__(self) -> int: return len(self.value)
    def __iter__(self) -> Iterable: return iter(self.value)
    def __reversed__(self) -> Iterable: return reversed(self.value)

    def __neg__(self): return self.value.__neg__()
    def __pos__(self): return self.value.__pos__()
    def __abs__(self): return self.value.__abs__()
    def __invert__(self): return self.value.__invert__()
    def __round__(self, ndigits=None): return self.value.__round__(ndigits)
    def __eq__(self, other): return self.value.__eq__(other)
    def __ne__(self, other): return self.value.__ne__(other)
    def __lt__(self, other): return self.value.__lt__(other)
    def __le__(self, other): return self.value.__le__(other)
    def __gt__(self, other): return self.value.__gt__(other)
    def __ge__(self, other): return self.value.__ge__(other)
    def __add__(self, other): return self.value.__add__(other)
    def __radd__(self, other): return self.value.__radd__(other)
    def __sub__(self, other): return self.value.__sub__(other)
    def __rsub__(self, other): return self.value.__rsub__(other)
    def __mul__(self, other): return self.value.__mul__(other)
    def __rmul__(self, other): return self.value.__rmul__(other)
    def __div__(self, other): return self.value.__div__(other)
    def __rdiv__(self, other): return self.value.__rdiv__(other)
    def __truediv__(self, other): return self.value.__truediv__(other)
    def __rtruediv__(self, other): return self.value.__rtruediv__(other)
    def __floordiv__(self, other): return self.value.__floordiv__(other)
    def __rfloordiv__(self, other): return self.value.__rfloordiv__(other)
    def __divmod__(self, other): return self.value.__divmod__(other)
    def __rdivmod__(self, other): return self.value.__rdivmod__(other)
    def __mod__(self, other): return self.value.__mod__(other)
    def __rmod__(self, other): return self.value.__rmod__(other)
    def __pow__(self, other): return self.value.__pow__(other)
    def __rpow__(self, other): return self.value.__rpow__(other)
    def __matmul__(self, other): return self.value.__matmul__(other)
    def __rmatmul__(self, other): return self.value.__rmatmul__(other)
    def __and__(self, other): return self.value.__and__(other)
    def __rand__(self, other): return self.value.__rand__(other)
    def __or__(self, other): return self.value.__or__(other)
    def __ror__(self, other): return self.value.__ror__(other)
    def __xor__(self, other): return self.value.__xor__(other)
    def __rxor__(self, other): return self.value.__rxor__(other)
    def __lshift__(self, other): return self.value.__lshift__(other)
    def __rlshift__(self, other): return self.value.__rlshift__(other)
    def __rshift__(self, other): return self.value.__rshift__(other)
    def __rrshift__(self, other): return self.value.__rrshift__(other)

    def __repr__(self) -> str:
        return self.tree_repr()

    def tree_repr(self, **kwargs) -> str:
        return tree_repr(self.value, **kwargs)

    def tree_flatten(self):
        return [self.value], None

    def tree_flatten_with_keys(self):
        return [('', self.value)], None

    @classmethod
    def tree_unflatten(cls, _, leaves):
        self = object.__new__(cls)
        self.value = leaves[0]

        return self


class Namespace(metaclass=PyTreeMeta):
    r"""PyTree class for name-value mappings.

    Arguments:
        kwargs: A name-value mapping.

    Example:
        >>> ns = Namespace(a=1, b='2'); ns
        Namespace(
          a = 1,
          b = '2'
        )
        >>> ns.c = [3, False]; ns
        Namespace(
          a = 1,
          b = '2',
          c = [3, False]
        )
        >>> jax.tree_util.tree_leaves(ns)
        [1, '2', 3, False]
    """

    def __init__(self, **kwargs):
        self.__dict__.update(**kwargs)

    def __repr__(self) -> str:
        return tree_repr(self)

    def tree_repr(self, **kwargs) -> str:
        lines = (
            f'{name} = {tree_repr(getattr(self, name), **kwargs)}'
            for name in sorted(self.__dict__.keys())
        )

        lines = ',\n'.join(lines)

        if lines:
            lines = '\n' + indent(lines, '  ') + '\n'

        return f'{self.__class__.__name__}({lines})'

    def tree_flatten(self):
        if self.__dict__:
            names, values = zip(*sorted(self.__dict__.items()))
        else:
            names, values = (), ()

        return values, names

    def tree_flatten_with_keys(self):
        values, names = self.tree_flatten()
        keys = map(jtu.GetAttrKey, names)

        return list(zip(keys, values)), names

    @classmethod
    def tree_unflatten(cls, names, values):
        self = object.__new__(cls)
        self.__dict__ = dict(zip(names, values))

        return self


class Static(metaclass=PyTreeMeta):
    r"""Wraps an hashable value as a leafless PyTree.

    Arguments:
        value: An hashable value to wrap.

    Example:
        >>> x = Static((0, 'one', None))
        >>> x.value
        (0, 'one', None)
        >>> jax.tree_util.tree_leaves(x)
        []
        >>> jax.tree_util.tree_structure(x)
        PyTreeDef(CustomNode(Static[(0, 'one', None)], []))
    """

    def __init__(self, value: Hashable):
        if not isinstance(value, Hashable):
            warn(f"'{type(value).__name__}' object is not hashable.")

        self.value = value

    def __hash__(self) -> int:
        return hash((type(self), self.value))

    def __repr__(self) -> str:
        return self.tree_repr()

    def tree_repr(self, **kwargs) -> str:
        return f'{self.__class__.__name__}({tree_repr(self.value, **kwargs)})'

    def tree_flatten(self):
        return (), self.value

    @classmethod
    def tree_unflatten(cls, value, _):
        self = object.__new__(cls)
        self.value = value

        return self


class Auto(Namespace):
    r"""Subclass of :class:`Namespace` that automatically detects non-array leaves
    and considers them as static.

    Important:
        :py:`object()` leaves are never considered static.

    Arguments:
        kwargs: A name-value mapping.

    Example:
        >>> auto = Auto(a=1, b=jnp.array(2.0)); auto
        Auto(
          a = 1,
          b = float32[]
        )
        >>> auto.c = ['3', jnp.arange(4)]; auto
        Auto(
          a = 1,
          b = float32[],
          c = ['3', int32[4]]
        )
        >>> jax.tree_util.tree_leaves(auto)  # only arrays
        [Array(2., dtype=float32, weak_type=True), Array([0, 1, 2, 3], dtype=int32)]
    """

    def tree_flatten(self):
        values, names = super().tree_flatten()

        values = jtu.tree_map(
            f=lambda x: x if type(x) is object or is_array(x) or is_auto(x) else Static(x),
            tree=values,
            is_leaf=is_auto,
        )

        return values, names

    @classmethod
    def tree_unflatten(cls, names, values):
        values = jtu.tree_map(
            f=lambda x: x.value if is_static(x) else x,
            tree=values,
            is_leaf=lambda x: is_auto(x) or is_static(x),
        )

        return super().tree_unflatten(names, values)


def tree_repr(
    x: PyTree,
    /,
    linewidth: int = 88,
    typeonly: bool = True,
    **kwargs,
) -> str:
    r"""Creates a pretty representation of a tree.

    Arguments:
        x: The tree to represent.
        linewidth: The maximum line width before elements of tuples, lists and dicts
            are represented on separate lines.
        typeonly: Whether to represent the type of arrays instead of their elements.

    Returns:
        The representation string.

    Example:
        >>> tree = [1, 'two', (True, False), list(range(5)), {'6': jnp.arange(7)}]
        >>> print(tree_repr(tree))
        [
          1,
          'two',
          (True, False),
          [0, 1, 2, 3, 4, 5],
          {'6': int32[7]}
        ]
    """

    kwargs.update(
        linewidth=linewidth,
        typeonly=typeonly,
    )

    if hasattr(x, 'tree_repr'):
        return x.tree_repr(**kwargs)
    elif isinstance(x, tuple):
        bra, ket = '(', ')'
        lines = [tree_repr(y, **kwargs) for y in x]
    elif isinstance(x, list):
        bra, ket = '[', ']'
        lines = [tree_repr(y, **kwargs) for y in x]
    elif isinstance(x, dict):
        bra, ket = '{', '}'
        lines = [
            f'{tree_repr(key)}: {tree_repr(value)}'
            for key, value in x.items()
        ]
    elif is_array(x):
        if typeonly:
            return f'{x.dtype}{list(x.shape)}'
        else:
            return repr(x)
    else:
        return repr(x).strip(' \n')

    if any('\n' in line for line in lines):
        lines = ',\n'.join(lines)
    elif sum(len(line) + 2 for line in lines) > linewidth:
        lines = ',\n'.join(lines)
    else:
        lines = ', '.join(lines)

    if '\n' in lines:
        lines = '\n' + indent(lines, '  ') + '\n'

    return f'{bra}{lines}{ket}'



r"""Tests for the inox.tree_util module."""

import jax
import pickle
import pytest

from jax import Array
from typing import *

from inox.tree_util import *


@pytest.mark.parametrize('nested', [False, True])
def test_PyArray(nested: bool):
    key = jax.random.key(0)
    x = jax.random.normal(key, (2, 3))

    if nested:
        y = PyArray(PyArray(x))
    else:
        y = PyArray(x)

    assert x.shape == y.shape
    assert x.dtype == y.dtype
    assert jax.numpy.allclose(x, y)

    # Operations
    z = x + y
    z = y * y
    z = jax.numpy.mean(y)

    # Flatten
    leaves, treedef = jax.tree_util.tree_flatten(y)

    assert all(isinstance(leaf, Array) for leaf in leaves)
    assert isinstance(treedef, Hashable)

    z = jax.tree_util.tree_unflatten(treedef, leaves)

    assert jax.numpy.allclose(x, z)

    # Pickle
    data = pickle.dumps(y)
    z = pickle.loads(data)

    assert jax.numpy.allclose(x, z)

    # Print
    assert repr(y)


def test_Namespace():
    x = Namespace(a=1, b='2')

    assert x.a == 1
    assert x.b == '2'

    # Add
    x.c = True

    assert x.c == True

    # Modify
    x.a = Namespace(d=None)

    assert x.a.d is None

    # Delete
    del x.b

    assert not hasattr(x, 'b')

    # Flatten
    leaves, treedef = jax.tree_util.tree_flatten(x)

    assert isinstance(treedef, Hashable)

    jax.tree_util.tree_unflatten(treedef, leaves)

    # Pickle
    data = pickle.dumps(x)
    pickle.loads(data)

    # Print
    assert repr(x)


@pytest.mark.parametrize('nested', [False, True])
def test_Static(nested: bool):
    x = 'hashable'

    if nested:
        y = Static(Static(x))
    else:
        y = Static(x)

    # Flatten
    leaves, treedef = jax.tree_util.tree_flatten(y)

    assert not leaves
    assert isinstance(treedef, Hashable)

    z = jax.tree_util.tree_unflatten(treedef, leaves)

    assert hash(y) == hash(z)

    # Pickle
    data = pickle.dumps(y)
    z = pickle.loads(data)

    assert hash(y) == hash(z)

    # Print
    assert repr(y)


def test_Auto():
    x = Auto(
        a=jax.numpy.ones(()),
        b=2,
        c=[True, jax.numpy.arange(4)],
    )

    # Flatten
    leaves, treedef = jax.tree_util.tree_flatten(x)

    assert all(isinstance(leaf, Array) for leaf in leaves)
    assert isinstance(treedef, Hashable)

    jax.tree_util.tree_unflatten(treedef, leaves)

    # Pickle
    data = pickle.dumps(x)
    pickle.loads(data)

    # Print
    assert repr(x)
